// functions that perform various smoothing procedures from the
// biomechanics literature
// Mode = 1 : Anderssen-Bloomfield regularization method with odd extension
// Mode = 2 : Anderssen-Bloomfield regularization method w/o odd extension 
// Mode = 3 : Kosarev-Pantos window method with odd extension
// Mode = 4 : Kosarev-Pantos window method w/o odd extension

#include "ABKP.H"


void mfilter(int mode, int nin, double dt, double x[SIZE2],
             double smsig[SIZE2])
{

   // declare some variables
   double der1,der2,s,xx,xy = 0.0;
   int n,m=0;
   double *lopt = new double[nin];
   
   // solobev norm is 2
   m = 2;
   
   double x1 = x[0];
   double slope = (x[nin-1] - x[0]) / (double)(nin-1);

   if ((mode == 1) || (mode == 3))
      n = 2 * nin - 1;
   else
      n = nin;
   
   int nm1 = n + 1;
   int mp1 = m - 1;
   int n2 = (int)(n / 2) + 1;
   double *gunorm = new double[n2];


   double omega = (2.0 * M_PI) / (nm1 * dt);
   Complex *coext = new Complex[2*nin]; 
   
   
   for (int i = 0; i < nin; ++i) {
      double rr = x[i] - x1 - (slope*i);
      coext[i] = Complex(rr,0);
      if ((mode == 1) || (mode == 3)) {
	 int ii = 2 * nin - (i+1);
         rr = -coext[i].real();
         coext[ii] = Complex(rr,0);
      }
              
   }
  

   // now compute the fft
   // call fft routine here, used a mixed radix approach if possible
   Complex *coext2 = new Complex[2*nin]; 
   dft(coext,nm1,coext2);   

   for (int i = 0; i < 2*nin; ++i)
      coext[i] = coext2[i];

   double *iv = new double[n2]; // holds periodogram data

   for (int i = 0; i < n2; ++i) {
      coext[i] = coext[i].conjg() / Complex(nm1,0.0); 
      iv[i] = (double)nm1 * coext[i].magnitude()*coext[i].magnitude() 
	      / (2.0*M_PI);
   }
   

   if ((mode == 3) || (mode == 4)) {
      // compute Kosarev-Pantos approx to Weiner window
      int nendsp = (int)(0.5 * (double)n2);
      double s = 0;
      // compute noise spectrum from nendsp to n2
      for (int i = nendsp-1; i < n2; ++i)
         s += iv[i];
      double noise = s / (n2 - nendsp + 1);

      int j0 = 2;
      // compute j0 value
      for (int i = 2; i <= nendsp; ++i) {
	 double s0 = iv[i];
         double s1 = iv[i+1];
         double s2 = iv[i+2];
         double s3 = iv[i+3];
         if ((s0 < noise) && ((s1 < noise) || (s2 < noise) ||
                              (s3 < noise))) {
	    j0 = i; 
            break;
	 }
      }
   
      // compute straightline extrapolation to log(iv)
      double xy,xx = 0.0;
      s = 0.0;
      for (int i = 2; i <= j0; ++i) {
	 xy += (i * log(iv[i]));    
         xx += (i*i);
         s += log(iv[i]);
      }

      double xm = (2.0 + j0) / 2.0;
      double ym = s / j0;
      double a1 = (xy - (j0*xm*ym)) / (xx - (j0*xm*xm));
      double b1 = ym - (a1 *xm);
      int j1 = j0;
         
      for (int i = j0; i < n2; ++i) 
	 if (exp((a1 * i) + b1)/noise < 0.01) {
	    j1 = i;
            break;
	 }
       
      // compute filter window in three frequency ranges
      for (int i = 0; i < j0; ++i)
         lopt[i] = iv[i] / (iv[i] + noise);
      for (int i = j0; i < j1; ++i)
         lopt[i] = exp((a1*i) + b1)/(exp((a1*i) + b1) + noise);
      for (int i = j1; i < n2; ++i)
         lopt[i] = 0.0;
   }
   else { // mode is either 1 or 2
      int id = 2;

      // optimize  f(alpha) to compute regularization param
      double t = nm1 + dt;
      int mi = m + id;
      gunorm[0] = 0;
      for (int i = 1; i < n2; ++i) {
	 double omj = 2.0 * M_PI * i;
         gunorm[i] = (pow(omj,-2*id) * ((omj*omj)-1.0)) / 
	             (pow(omj,2*mp1) - 1.0);
      }
      double fumin = 1e30;
      double alfaa = 0.001;
      double alfami = alfaa;
      double fun,d1,d2 = 0.0;
       

      for (int i = 0; i < 50; ++i) {
	 alfaa /= 3.0;
         f_alpha(iv,gunorm,n2,alfaa,&fun,&d1,&d2,1);
         if (fun < fumin) {
	    alfami = alfaa;
            fumin = fun;
         }
      }
      
      int itopt = 0;
      alfaa = alfami;
      fun = fumin;
      // refine raw minumum by newton-raphson search
      double oldfun = 0;
      for (;;) {
	 oldfun = fun;
         f_alpha(iv,gunorm,n2,alfaa,&fun,&d1,&d2,0);
	 double step = -d1 / d2;
      	 if ((fun > oldfun) || (itopt > 7) || (fabs(step <= alfaa*1e-6)))
	    break;
         ++itopt;
         alfaa += step;
      }
     
      double alfa = alfaa;
      lopt[0] = 1.0;
      double bopt = 0.0;
      for (int i = 1; i < n2; ++i) {
	 lopt[i] = gunorm[i] / (gunorm[i] + alfa);
         double gv = 1.0 + gunorm[i] / alfa;
         double t = iv[i] / gv;
         bopt += t;  
      }
      bopt /= (double)(n2-1);
   }
 

   Complex *dfts = new Complex[2*nin];
   Complex z(1,0);
   // now finish it up
   for (int i = 1; i < n2; ++i) {
      Complex lcom(lopt[i],0);
      dfts[i] = coext[i] * lcom * z;
      int jj = n - i + 1; 
      dfts[jj] = dfts[i].conjg();
   }
   
    
   // call fft on dfts here
   Complex *dfts2 = new Complex[2*nin];
   dft(dfts,nm1,dfts2);

   for (int i = 0; i < 2*nin; ++i) {
      dfts[i] = dfts2[i];
   }
   
   dfts[n-1] = Complex(dfts[0].real(), dfts[n-1].imag());
    
   for (int i = 0; i < nin; ++i)
      smsig[i] = dfts[i].real() + x1 + (slope * i);

   // clean up
   delete [] lopt;
   delete [] gunorm;
   delete [] iv; 
   delete [] coext;
   delete [] coext2;
   delete [] dfts;
   delete [] dfts2;

}


void f_alpha(double iv[SIZE2], double gunorm[SIZE2], int n, double alpha,
             double *d0, double *d1, double *d2, int ind)
{
   int r1 = n - 1;
   double sfa = 0.0;
   double sfa2 = 0.0;
   double sfab = 0.0;
   double sfa2b = 0.0;
   double sfa3b = 0.0;
   double slog = 0.0;

   for (int i = 1; i < n; ++i) {
      double fa = gunorm[i] / (1.0 + alpha * gunorm[i]);
      // double fa = 1.0 / (gunorm[i] + alpha);
      double fab = iv[i] * fa;
      sfab += fab;
      slog += log(fa);
      if (ind == 0) {
	 double fa2 = fa * fa;
         double fa3 = fa2 * fa;
         double fa2b = iv[i] * fa2;
         double fa3b = iv[i] * fa3;
         sfa += fa;
         sfa2 +=fa2;
         sfa2b += fa2b;
         sfa3b += fa3b;
      } 
   }
      *d0 = r1 * log(sfab) - slog;
      *d1 = r1 * (-sfa2b / sfab) + sfa;
      *d2 = r1 * ((sfab * 2.0 * sfa3b - sfa2b*sfa2b) / (sfab*sfab)) - sfa2;
}


void dft(Complex x[SIZE2], int n, Complex y[SIZE2])
{
   for (int k = 0; k < n; ++k) {
      y[k] = Complex(0,0);
      for (int j = 0; j < n; ++j) 
	 y[k] += Complex(x[j].real() * cos(2.0*M_PI*k*j/n) - 
                         x[j].imag() * sin(2.0*M_PI*k*j/n),
                         x[j].imag() * cos(2.0*M_PI*k*j/n) +
                         x[j].real() * sin(2.0*M_PI*k*j/n)); 
   }
}
