#include "gauss.H"
#include <iostream.h>

GaussWhiteNoise::GaussWhiteNoise(unsigned int seed)
{
  srandom(seed);

  _sampleRate   = 0;
  _duration     = 0;
  _last_fds     = 0;
  _last_fds2    = 0;
  _last_fds3    = 0;

  _noise        = NULL;
  _noiseFirst   = NULL;
  _noiseCurrent = NULL;
  _noiseLast    = NULL;

  _noiseFirst2   = NULL;
  _noiseCurrent2 = NULL;
  _noiseLast2    = NULL;

  _noiseFirst3   = NULL;
  _noiseCurrent3 = NULL;
  _noiseLast3    = NULL;
  

}

GaussWhiteNoise::GaussWhiteNoise(char* noise_file, unsigned int seed, int verbose)
{
  double cutoff = 0;

  // noise data file
  FILE *fp = fopen(noise_file, "r");
  if(fp == NULL) {
    perror(noise_file);
    exit(1);
  }

  // Attempt to read noise file header which contains
  // duration   duration of the noise signal in seconds
  // sampleRate sample rate in Hz
  // cutoff     cutoff frequency in Hz
  if(fscanf(fp, "%lf %lf %lf",&_duration,&_sampleRate,&cutoff) == EOF) {
    fprintf(stderr, "ERROR: Error reading noise file header\n");
    exit(-1);
  }

  // print status
  if (verbose) {
    fprintf(stderr,"[GaussWhiteNoise] reading file %s,\n",noise_file);
    fprintf(stderr,"[GaussWhiteNoise] %g seconds of %g Hz bandlimited noise sampled at %g Hz...\n",_duration,cutoff,_sampleRate);
  }

  int samples  = (int)(_duration*_sampleRate);
  _noise        = new double[samples];
  _noiseCurrent = &_noise[0];
  _noiseFirst   = &_noise[0];
  _noiseLast    = &_noise[samples-1];
  for (int s=0; s<samples; s++) {
    if (fscanf(fp, "%lf\n",_noiseCurrent++) != 1) {
      fprintf(stderr, "[GaussWhiteNoise] ERROR: bad line %i in noise file\n",s+1);
      exit(-1);
    }
  }
  if (verbose) fprintf(stderr,"[GaussWhiteNoise] done.\n");

  // seed random number gen
  // randomly choose a starting position
  srand48((unsigned long)seed);
  int ndx = (int)(lrand48() % (long)samples);
  _noiseCurrent = &_noise[ndx];
  _last_fds = 1.0;
}

GaussWhiteNoise::~GaussWhiteNoise()
{
  deleteNoise();
}


void GaussWhiteNoise::deleteNoise() 
{
  if (_noise != NULL)
   delete [] _noise;

}

void GaussWhiteNoise::generateNoise(double srate, double duration, 
                                    double variance, int cap) 
{
 _var = variance;
 _sampleRate = srate;
 _duration = duration;

 if (_noise != NULL)
   deleteNoise(); 

 // generate noise for the duration time at the given sampling rate
  int samples   = (int)(_duration*_sampleRate);
  _noise        = new double[samples];
  resetCounters();

  for (int s=0; s<samples; s++) {
    double sp = sample();
    if (cap) {
      if (sp > _var)
	sp = _var;
      if (sp < -_var)
	sp = -_var;
    }
    _noise[s] = sp; // take from Normal dist, mean 0, given var

  }

}

void GaussWhiteNoise::resetCounters() 
{
  if (_noise == NULL)
    return;

   int samples   = (int)(_duration*_sampleRate);
  _noiseCurrent = _noiseCurrent2 = _noiseCurrent3 = &_noise[0];
  _noiseFirst   = _noiseFirst2   = _noiseFirst3   = &_noise[0];
  _noiseLast    = _noiseLast2    = _noiseLast3    = &_noise[samples-1];
}

void GaussWhiteNoise::randomizeCounters(unsigned int seed)
{
  if (_noise == NULL)
   return;

  int samples  = (int)(_duration*_sampleRate);
 // seed random number gen
  // randomly choose a starting position
  srand48((unsigned long)seed);
  int ndx = (int)(lrand48() % (long)samples);
  _noiseCurrent = &_noise[ndx];
  _last_fds = 1.0;
  
  ndx = (int)(lrand48() % (long)samples);
  _noiseCurrent2 = &_noise[ndx];
  _last_fds2 = 1.0;
  
  ndx = (int)(lrand48() % (long)samples);
  _noiseCurrent3 = &_noise[ndx];
  _last_fds3 = 1.0;

}


double GaussWhiteNoise::sample()
{
   double dRandA = erand48(_xsubi);
   double dRandB = 2.0*((double)M_PI)*erand48(_xsubi);

   double dRay = pow((2.0*_var*log(1.0/(1.0-dRandA))), 0.5);

   return(dRay * cos(dRandB));
}

double GaussWhiteNoise::sample(double dt)
{
  // make sure properly initialized
  if (_sampleRate == 0) {
    fprintf(stderr, "[GaussWhiteNoise] ERROR: uninitialized noise signal.\n");
    exit(-1);
  }

  // compute how many samples for dt
  double  ds = dt * _sampleRate; // real number of samples
  int    ids = (int)ds;  // integer part
  double fds = ds - (double)ids; // fractional part

  // add remainder of fds from last time and see if causes to rollover
  fds += (1.0 - _last_fds);
  if (fds >= 1.0) {
    ids += 1;
    fds -= 1.0;
  }
  _last_fds = fds;

  double thisSample = 0;
  double nextSample = 0;
 
  // advance noise pointer for intermediate signal values (-1 to interpolate)
  _noiseCurrent += (ids-1)*sizeof(double);

  if (_noiseCurrent>_noiseLast) {
    unsigned long over = _noiseCurrent - _noiseLast;
    _noiseCurrent = _noiseFirst + over - sizeof(double);
  }

  // last whole sample
  thisSample = *_noiseCurrent;

  // if there is some fractional part
  if (fds>0) {
    _noiseCurrent++;
    if (_noiseCurrent>_noiseLast) _noiseCurrent = _noiseFirst;
    nextSample = *_noiseCurrent;

    // interpolate between last and next
    double slope = (nextSample - thisSample) * _sampleRate;
    thisSample += slope*(fds/_sampleRate);
  }

  // return random noise signal sample
  return(thisSample);
}


double GaussWhiteNoise::sample(double dt, int i)
{

 // make sure properly initialized
  if (_noise == NULL) {
    fprintf(stderr, "[GaussWhiteNoise] ERROR: uninitialized noise signal.\n");
    return 0.0;
  }

  // compute how many samples for dt
  double  ds = dt * _sampleRate; // real number of samples
  int    ids = (int)ds;  // integer part
  double fds = ds - (double)ids; // fractional part

  // add remainder of fds from last time and see if causes to rollover
  double tmpfds = 0.0;
  if (i == 1)
    tmpfds = _last_fds;
  else if (i == 2)
    tmpfds = _last_fds2;
  else if (i == 3) 
    tmpfds = _last_fds3;

  fds += (1.0 - tmpfds);
  if (fds >= 1.0) {
    ids += 1;
    fds -= 1.0;
  }

  if (i == 1)
   _last_fds = fds;
  else if (i == 2)
   _last_fds2 = fds;
  else if (i == 3) 
   _last_fds3 = fds;

  double thisSample = 0;
  double nextSample = 0;
 
  // advance noise pointer for intermediate signal values (-1 to interpolate)
  if (i == 1) {
    _noiseCurrent += (ids-1)*sizeof(double);

    if (_noiseCurrent > _noiseLast) {
       unsigned long over = _noiseCurrent - _noiseLast;
      _noiseCurrent = _noiseFirst + over - sizeof(double);
     }

    // last whole sample
    thisSample = *_noiseCurrent;
    //     cerr << "Over 1 Before" << endl;
    // if there is some fractional part
    if (fds>0) {
      _noiseCurrent++;
      if (_noiseCurrent>_noiseLast) 
	_noiseCurrent = _noiseFirst;
      nextSample = *_noiseCurrent;

      // interpolate between last and next
      double slope = (nextSample - thisSample) * _sampleRate;
      thisSample += slope*(fds/_sampleRate);
    }
    //  cerr << "Over 1 After" << endl;
  }
  else if (i == 2) {
   _noiseCurrent2 += (ids-1)*sizeof(double);

    if (_noiseCurrent2>_noiseLast2) {
      unsigned long over = _noiseCurrent2 - _noiseLast2;
      _noiseCurrent2 = _noiseFirst2 + over - sizeof(double);
    }

    // last whole sample
    thisSample = *_noiseCurrent2;
    //  cerr << "Over 2 Before" << endl;
    // if there is some fractional part
    if (fds>0) {
      _noiseCurrent2++;
      if (_noiseCurrent2>_noiseLast2) 
	_noiseCurrent2 = _noiseFirst2;
      nextSample = *_noiseCurrent2;
      
      // interpolate between last and next
      double slope = (nextSample - thisSample) * _sampleRate;
      thisSample += slope*(fds/_sampleRate);
    }
    //  cerr << "Over 2 After" << endl; 
  }
  else if (i == 3) {
   _noiseCurrent3 += (ids-1)*sizeof(double);

    if (_noiseCurrent3>_noiseLast3) {
      unsigned long over = _noiseCurrent3 - _noiseLast3;
      _noiseCurrent3 = _noiseFirst3 + over - sizeof(double);
    }

    // last whole sample
    thisSample = *_noiseCurrent3;
    //    cerr << "Over 3 Before" << endl;
    // if there is some fractional part
    if (fds>0) {
      _noiseCurrent3++;
      if (_noiseCurrent3>_noiseLast3) 
	_noiseCurrent3 = _noiseFirst3;
      nextSample = *_noiseCurrent3;
      
      // interpolate between last and next
      double slope = (nextSample - thisSample) * _sampleRate;
      thisSample += slope*(fds/_sampleRate);
    }
    //  cerr << "Over 3 After" << endl;
  }

  // return random noise signal sample
  return(thisSample);


}
