// File: pose.C
// Contains methods for reading and writing to disk

#include "pose.H"
#include <stdio.h>
#include <stdlib.h>
#include "ui.H"

// loads in data from file using simple data file format

void UMotionRec::loadSimpleData(char *file)
{
 FILE *fp;

 if ((fp = fopen(file, "r")) == NULL) {
   cerr << "Bad input file name!" << endl;
   exit(0);
 }

 // read in sampling rate
 fscanf(fp, "%lf\n", &_srate);
 // read records and store // assuming velocity and accel info are zeros
 while (!feof(fp)) {
   double t,x,y,z,qx,qy,qz,qw;
   fscanf(fp, "%lf %lf %lf %lf %lf %lf %lf %lf\n", &t,&x,&y,&z,&qx,&qy,&qz,&qw);
   Pose *ps = new Pose(t, Vec3d(x,y,z), Vec3d(0.0,0.0,0.0), Vec3d(0.0,0.0,0.0),
                       Quat(qx,qy,qz,qw), Quat(0.0,0.0,0.0,0.0), Quat(0.0,0.0,0.0,0.0));
   _poselist += ps;

 }
 
 fclose(fp);   
}


// loads im data from file using regular format

void UMotionRec::loadData(const char *file)
{
  FILE *fp;
  
  if ((fp = fopen(file, "r")) == NULL) {
    cerr << "Bad input file name!" << endl;
    exit(0);
  }

  // read in sampling rate
  fscanf(fp, "%lf\n", &_srate);
  // read records and store data in the pose list
  while (!feof(fp)) {
    double t,x,y,z,xv,yv,zv,xa,ya,za;
    double qx,qy,qz,qw,avx,avy,avz,avw,aax,aay,aaz,aaw;
    fscanf(fp, "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf\n", 
	   &t,&x,&y,&z,&xv,&yv,&zv,&xa,&ya,&za,&qx,&qy,&qz,&qw,&avx,&avy,&avz,&avw,&aax,&aay,&aaz,&aaw);
    Pose *ps = new Pose(t, Vec3d(x,y,z), Vec3d(xv,yv,zv), Vec3d(xa,ya,za),
			Quat(qx,qy,qz,qw), Quat(avx,avy,avz,avw), Quat(aax,aay,aaz,aaw));
    _poselist += ps;
    
  }
  
  fclose(fp);   
}


// save data to disk

void UMotionRec::saveData(const char *file)
{
  FILE *fp;

  if ((fp = fopen(file, "w")) == NULL) {
    cerr << "Cannot write to file!" << endl;
    exit(0);
  }

  // write out the sampling rate
  fprintf(fp, "%lf\n", _srate);

  for (int i = 0; i < _poselist.num(); ++i) 
    fprintf(fp, "%3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf %3.12lf\n",
            _poselist[i]->_tstamp,
            _poselist[i]->_pos[0],_poselist[i]->_pos[1],_poselist[i]->_pos[2],
            _poselist[i]->_posvel[0],_poselist[i]->_posvel[1],_poselist[i]->_posvel[2],
            _poselist[i]->_posacc[0],_poselist[i]->_posacc[1],_poselist[i]->_posacc[2], 
            _poselist[i]->_orient.V()[0], _poselist[i]->_orient.V()[1],_poselist[i]->_orient.V()[2],
            _poselist[i]->_orient.S(),
	    _poselist[i]->_angvel.V()[0], _poselist[i]->_angvel.V()[1],_poselist[i]->_angvel.V()[2],
            _poselist[i]->_angvel.S(),
	    _poselist[i]->_angacc.V()[0], _poselist[i]->_angacc.V()[1],_poselist[i]->_angacc.V()[2],
            _poselist[i]->_angacc.S());

  fclose(fp);
}


void UMotionRec::sendToMatlab(MATLABEngine *eng)
{
  // first clear out the old matlab variables
  eng->runCmd(QString("clear"));

  // send the data as a set of arrays to the matlab engine
  mxArray *T = NULL;
  mxArray *X = NULL;
  mxArray *Y = NULL;
  mxArray *Z = NULL;
  mxArray *VX = NULL;
  mxArray *VY = NULL;
  mxArray *VZ = NULL;
  mxArray *AX = NULL;
  mxArray *AY = NULL;
  mxArray *AZ = NULL;
  mxArray *QX = NULL;
  mxArray *QY = NULL;
  mxArray *QZ = NULL;
  mxArray *QW = NULL;
  mxArray *VQX = NULL;
  mxArray *VQY = NULL;
  mxArray *VQZ = NULL;
  mxArray *VQW = NULL;
  mxArray *AQX = NULL;
  mxArray *AQY = NULL;
  mxArray *AQZ = NULL;
  mxArray *AQW = NULL;

  int num = _poselist.num(); 

  double *t = new double[num];
  double *x = new double[num];
  double *y = new double[num];
  double *z = new double[num];
  double *vx = new double[num];
  double *vy = new double[num];
  double *vz = new double[num];
  double *ax = new double[num];
  double *ay = new double[num];
  double *az = new double[num];
  double *qx = new double[num];
  double *qy = new double[num];
  double *qz = new double[num];
  double *qw = new double[num];
  double *vqx = new double[num];
  double *vqy = new double[num];
  double *vqz = new double[num];
  double *vqw = new double[num];
  double *aqx = new double[num];
  double *aqy = new double[num];
  double *aqz = new double[num];
  double *aqw = new double[num];

  for (int i = 0; i < num; ++i) {
    t[i] = _poselist[i]->_tstamp;
    x[i] = _poselist[i]->_pos[0];
    y[i] = _poselist[i]->_pos[1];
    z[i] = _poselist[i]->_pos[2]; 
    vx[i] = _poselist[i]->_posvel[0];
    vy[i] = _poselist[i]->_posvel[1];
    vz[i] = _poselist[i]->_posvel[2]; 
    ax[i] = _poselist[i]->_posacc[0];
    ay[i] = _poselist[i]->_posacc[1];
    az[i] = _poselist[i]->_posacc[2]; 
    qx[i] = _poselist[i]->_orient.V()[0];
    qy[i] = _poselist[i]->_orient.V()[1];
    qz[i] = _poselist[i]->_orient.V()[2]; 
    qw[i] = _poselist[i]->_orient.S();
    vqx[i] = _poselist[i]->_angvel.V()[0];
    vqy[i] = _poselist[i]->_angvel.V()[1];
    vqz[i] = _poselist[i]->_angvel.V()[2]; 
    aqw[i] = _poselist[i]->_angacc.S();
    aqx[i] = _poselist[i]->_angacc.V()[0];
    aqy[i] = _poselist[i]->_angacc.V()[1];
    aqz[i] = _poselist[i]->_angacc.V()[2]; 
    aqw[i] = _poselist[i]->_angacc.S();
  }
  
  T = mxCreateDoubleMatrix(1, num, mxREAL);
  X = mxCreateDoubleMatrix(1, num, mxREAL);
  Y = mxCreateDoubleMatrix(1, num, mxREAL);
  Z = mxCreateDoubleMatrix(1, num, mxREAL);
  VX = mxCreateDoubleMatrix(1, num, mxREAL);
  VY = mxCreateDoubleMatrix(1, num, mxREAL);
  VZ = mxCreateDoubleMatrix(1, num, mxREAL);
  AX = mxCreateDoubleMatrix(1, num, mxREAL);
  AY = mxCreateDoubleMatrix(1, num, mxREAL);
  AZ = mxCreateDoubleMatrix(1, num, mxREAL);
  QX = mxCreateDoubleMatrix(1, num, mxREAL);
  QY = mxCreateDoubleMatrix(1, num, mxREAL);
  QZ = mxCreateDoubleMatrix(1, num, mxREAL);
  QW = mxCreateDoubleMatrix(1, num, mxREAL);
  VQX = mxCreateDoubleMatrix(1, num, mxREAL);
  VQY = mxCreateDoubleMatrix(1, num, mxREAL);
  VQZ = mxCreateDoubleMatrix(1, num, mxREAL);
  VQW = mxCreateDoubleMatrix(1, num, mxREAL);
  AQX = mxCreateDoubleMatrix(1, num, mxREAL);
  AQY = mxCreateDoubleMatrix(1, num, mxREAL);
  AQZ = mxCreateDoubleMatrix(1, num, mxREAL);
  AQW = mxCreateDoubleMatrix(1, num, mxREAL);

  mxSetName(T,"T");
  mxSetName(X,"X");
  mxSetName(Y,"Y");
  mxSetName(Z,"Z");
  mxSetName(VX,"VX");
  mxSetName(VY,"VY");
  mxSetName(VZ,"VZ");
  mxSetName(AX,"AX");
  mxSetName(AY,"AY");
  mxSetName(AZ,"AZ");
  mxSetName(QX,"QX");
  mxSetName(QY,"QY");
  mxSetName(QZ,"QZ");
  mxSetName(QW,"QW");
  mxSetName(VQX,"VQX");
  mxSetName(VQY,"VQY");
  mxSetName(VQZ,"VQZ");
  mxSetName(VQW,"VQW");
  mxSetName(AQX,"AQX");
  mxSetName(AQY,"AQY");
  mxSetName(AQZ,"AQZ");
  mxSetName(AQW,"AQW");
 
  memcpy((void*)mxGetPr(T), (void*)t, num*sizeof(double));
  memcpy((void*)mxGetPr(X), (void*)x, num*sizeof(double));
  memcpy((void*)mxGetPr(Y), (void*)y, num*sizeof(double));
  memcpy((void*)mxGetPr(Z), (void*)z, num*sizeof(double));
  memcpy((void*)mxGetPr(VX), (void*)vx, num*sizeof(double));
  memcpy((void*)mxGetPr(VY), (void*)vy, num*sizeof(double));
  memcpy((void*)mxGetPr(VZ), (void*)vz, num*sizeof(double));  
  memcpy((void*)mxGetPr(AX), (void*)ax, num*sizeof(double));
  memcpy((void*)mxGetPr(AY), (void*)ay, num*sizeof(double));
  memcpy((void*)mxGetPr(AZ), (void*)az, num*sizeof(double));  
  memcpy((void*)mxGetPr(QX), (void*)qx, num*sizeof(double));  
  memcpy((void*)mxGetPr(QY), (void*)qy, num*sizeof(double));
  memcpy((void*)mxGetPr(QZ), (void*)qz, num*sizeof(double));
  memcpy((void*)mxGetPr(QW), (void*)qw, num*sizeof(double));  
  memcpy((void*)mxGetPr(VQX), (void*)vqx, num*sizeof(double));  
  memcpy((void*)mxGetPr(VQY), (void*)vqy, num*sizeof(double));
  memcpy((void*)mxGetPr(VQZ), (void*)vqz, num*sizeof(double));
  memcpy((void*)mxGetPr(VQW), (void*)vqw, num*sizeof(double));  
  memcpy((void*)mxGetPr(AQX), (void*)aqx, num*sizeof(double));  
  memcpy((void*)mxGetPr(AQY), (void*)aqy, num*sizeof(double));
  memcpy((void*)mxGetPr(AQZ), (void*)aqz, num*sizeof(double));
  memcpy((void*)mxGetPr(AQW), (void*)aqw, num*sizeof(double));   

  eng->sendArray(T); 
  eng->sendArray(X); 
  eng->sendArray(Y); 
  eng->sendArray(Z); 
  eng->sendArray(VX); 
  eng->sendArray(VY); 
  eng->sendArray(VZ); 
  eng->sendArray(AX); 
  eng->sendArray(AY); 
  eng->sendArray(AZ);
  eng->sendArray(QX); 
  eng->sendArray(QY); 
  eng->sendArray(QZ); 
  eng->sendArray(QW);
  eng->sendArray(VQX); 
  eng->sendArray(VQY); 
  eng->sendArray(VQZ); 
  eng->sendArray(VQW);
  eng->sendArray(AQX); 
  eng->sendArray(AQY); 
  eng->sendArray(AQZ); 
  eng->sendArray(AQW);

  // free up this memory
  mxDestroyArray(T);
  mxDestroyArray(X);
  mxDestroyArray(Y);
  mxDestroyArray(Z);
  mxDestroyArray(VX);
  mxDestroyArray(VY);
  mxDestroyArray(VZ);
  mxDestroyArray(AX);
  mxDestroyArray(AY);
  mxDestroyArray(AZ);
  mxDestroyArray(QX);
  mxDestroyArray(QY);
  mxDestroyArray(QZ);
  mxDestroyArray(QW);
  mxDestroyArray(VQX);
  mxDestroyArray(VQY);
  mxDestroyArray(VQZ);
  mxDestroyArray(VQW); 
  mxDestroyArray(AQX);
  mxDestroyArray(AQY);
  mxDestroyArray(AQZ);
  mxDestroyArray(AQW); 

  delete [] t;
  delete [] x;  
  delete [] y;
  delete [] z;  
  delete [] vx;  
  delete [] vy;
  delete [] vz;    
  delete [] ax;  
  delete [] ay;
  delete [] az;      
  delete [] qx;    
  delete [] qy;  
  delete [] qz;
  delete [] qw;    
  delete [] vqx;    
  delete [] vqy;  
  delete [] vqz;
  delete [] vqw;  
  delete [] aqx;    
  delete [] aqy;  
  delete [] aqz;
  delete [] aqw;  
}

  

