// functions will do general covariance intersection under KF and EKF equations

#include "cibase.H"
#include "ciomega.H"

// standard CI-KF
void
CIKalmanBase::ciKalmanFilter(Pose *userpos, int zNum, KALMANMODEL mod)
{
   // do the time update step
   MMatrix pstate = _fundMat * _state;

   // do the a priori extimate of the the process covariance matrix
   MMatrix pcov = _fundMat * (_PP * transp(_fundMat)) + _QQ;

   // find omega
   double omega = 1.0;
   MMatrix A(6,6);
   MMatrix B(3,3);
   switch(mod) {
     case kPVSM:
       omega = ComputeOmegaCI(pcov, A, 6, _RR, B, 3, _HH);  
       break;
     case kPVMMV:
       omega = ComputeOmegaCI(pcov, A, 6, _RR, A, 6, _HH);  
       break;
     default: ;
   }

   MMatrix tmp = (1.0 - omega) * _HH * (pcov * transp(_HH)) + (omega * _RR);
   MMatrix KG = (1.0 - omega) * pcov * (transp(_HH) * minverse2(tmp));
 
   // do the measurement update step
   MMatrix Z(zNum,1);
   switch(mod) {
     case kPVSM:
       Z(0,0) = userpos->_pos[0];
       Z(1,0) = userpos->_pos[1];
       Z(2,0) = userpos->_pos[2];
       break;
     case kPVMMV:
       Z(0,0) = userpos->_pos[0];
       Z(1,0) = userpos->_pos[1];
       Z(2,0) = userpos->_pos[2];
       Z(3,0) = userpos->_posvel[0];
       Z(4,0) = userpos->_posvel[1];
       Z(5,0) = userpos->_posvel[2];
       break;
     default: ;
   }

// 2 other CI variations     
//   MMatrix pcovi = minverse2((1.0 / omega) * pcov);
//   MMatrix Ri = minverse2((1.0 / (1.0 - omega)) * _RR);
//   _PP = minverse2(pcovi + (transp(_HH) * Ri * _HH));
//   _state = _PP * ((pcovi * pstate) + (transp(_HH) * (Ri * Z)));
  
//    MMatrix pcovi = minverse2(pcov);
//    MMatrix Ri = minverse2(_RR);
//    _PP = minverse2(omega*pcovi + (1.0- omega) * (transp(_HH) * Ri * _HH));
//    MMatrix nu = Z - _HH * pstate;
//    MMatrix W = (1.0 - omega) * _PP * transp(_HH) * Ri;
//    _state = pstate + W * nu;
 
    _state = pstate + KG * (Z - (_HH * pstate));   
    _PP = (1.0 / omega )*(pcov - (KG * _HH * pcov));

    
//    cerr << " " << _state(0,0) << " " << sqrt(_PP(0,0)) << endl;
}


// CI-EKF Filter
void
CIExtendedKalmanBase::ciExtendedKalmanFilter(Pose *userpos, int zNum, KALMANMODEL mod)
{
  Vec3d v(_Estate(0,0), _Estate(1,0), _Estate(2,0)); 

   // do the time update step using _edt and _Estate
   MMatrix epstate = _up->RK4(_Estate, _edt, mod, length(v));
   
   // update the fundamental matrix, Q matrix
   _EfundMat = _up->updateFundMat(epstate, _edt, mod, length(v));
   _EQ = _up->updateQMat(epstate, _edt, _spd, mod, length(v));

   // do the a priori estimate of the covariance matrix
   MMatrix epcov = _EfundMat * (_EP * transp(_EfundMat)) + _EQ;

   // find omega	
   double omega = 1.0;
   MMatrix A(6,6);
   MMatrix B(3,3);
   switch(mod) {
     case kPVSM:
       omega = ComputeOmegaCI(epcov, A, 6, _ER, B, 3, _EH);  
       break;
     case kPVMMV:
       omega = ComputeOmegaCI(epcov, A, 6, _ER, A, 6, _EH);  
       break;
     default: ;
     }
    
   MMatrix tmp = (1.0 - omega) * _EH * (epcov * transp(_EH)) + (omega * _ER);   
   MMatrix EKG = (1.0 - omega) * epcov * (transp(_EH) * minverse2(tmp));

   // do the mearurment update
   MMatrix EZ(zNum,1);
   switch(mod) {
    case kPVSM: {
       Vec3d tmpv = logMap(userpos->_orient);
       for (int i = 0; i < 3; ++i)
	  EZ(i,0) = tmpv[i];
      break;
    }
    case kPVMMV: {
       Vec3d tmpv = logMap(userpos->_orient);
       for (int i = 0; i < 3; ++i) {
	  EZ(i,0) = tmpv[i];
          EZ(i+3,0) = userpos->_angvel.V()[i];
       }
       break;
    }
     default: ;
   }

// Different approach    
//   MMatrix epcovi = minverse2(epcov);
//   MMatrix Ri = minverse2(_ER);
//   _EP = minverse2(omega*epcovi + (1.0- omega) * (transp(_EH) * Ri * _EH));
//   MMatrix nu = EZ - _EH * epstate;
//   MMatrix W = (1.0 - omega) * _EP * transp(_EH) * Ri;
//   _Estate = epstate + W * nu;	

   _Estate = epstate + EKG * (EZ - (_EH * epstate));   
   _EP = (1.0 / omega) * (epcov - (EKG * _EH * epcov));
 
   //   cerr << " " << _Estate(0,0) << " " << sqrt(_EP(0,0)) << endl;
}   
