#include "kalmanPVMMV.H"
#include "../pose.H"


// functions for the position/velocity multiple measurement (pos and vel)
// kalman predictor

// constuctor, create all variables according to PVMM specs
KalmanPVMMV::KalmanPVMMV() : KalmanBase(), ExtendedKalmanBase(), 
                             UnscentedKalmanBase(), _ctr(0), _ukfflag(0) 
{
 _RR = MMatrix(2,2);
 _QQ = MMatrix(2,2);
 _HH = MMatrix(2,2);
 _fundMat = MMatrix(2,2);

 for (int i = 0; i < 3; ++i) {
   _PP[i] = MMatrix(2,2);
   _state[i] = MMatrix(2,1);
 }

  // create for EKF
  _ER = MMatrix(7,7);
  _EH = MMatrix(7,7);
  _EQ = MMatrix(7,7);
  _EfundMat = MMatrix(7,7);
  _EP = MMatrix(7,7);
  _Estate = MMatrix(7,1);
  _edt = _spd = 0.0;

 // create for the UKF	
 _UR = MMatrix(7,7);
 _UQ = MMatrix(7,7);
 _UP = MMatrix(7,7);
 _Ustate = MMatrix(7,1);
 _udt = _uspd = 0.0;
 _dim = 7;  // 10 dimensional state vector

}

// destructor
KalmanPVMMV::~KalmanPVMMV()
{
  _RR.deleteMatrix();
  _HH.deleteMatrix();
  _QQ.deleteMatrix();
  _fundMat.deleteMatrix();
  for (int i = 0; i < 3; ++i) {
    _PP[i].deleteMatrix(); 
    _state[i].deleteMatrix();
  }

  _ER.deleteMatrix();
  _EH.deleteMatrix();
  _EQ.deleteMatrix();
  _EfundMat.deleteMatrix();
  _EP.deleteMatrix();
  _Estate.deleteMatrix();

  _UR.deleteMatrix();
  _UQ.deleteMatrix();
  _UP.deleteMatrix();
  _Ustate.deleteMatrix();
}


// initialize all variables according to PVMM specs
// except for state vectors
void
KalmanPVMMV::initialize(double spd, double dt, double np, double nq, 
                        double dv, int ukfflag) 
{
 _ctr = 0;
 _ukfflag = ukfflag; 

 _HH.setIdentity();
 _RR(0,0) = np;
 _RR(1,1) = sqrt(np);    //pow(dv,4) / 30.0;// assume error variance is no worse than
                   // the sqrt(error variance of position measurements)
                   // until something better comes along
 _fundMat(0,0) = _fundMat(1,1) = 1.0;
 _fundMat(0,1) = dt; 

 _QQ(0,0) = ((dt*dt*dt) / 3.0) * spd;
 _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
 _QQ(1,1) = dt * spd;

 for (int i = 0; i < 3; ++i) {
   _PP[i].setZeros();
   _PP[i](0,0) = _PP[i](1,1) = INITVAR;
   // zero out the state vectors
   _state[i](0,0) = _state[i](1,0) = 0.0;
 }


 if (!_ukfflag) {
   // initialize for EKF
   _edt = dt;
   _spd = spd;
   // we can set the measurment noise covariance matrix here (at least that can
   // stay constant
   _ER.setZeros();
   _EP.setZeros();
   for (int i = 0; i < 7; ++i) {
     if (i < 4) {
       _EP(i,i) = 1.0;
       _ER(i,i) = nq;
     }
     else {
       _EP(i,i) = INITVAR;
       _ER(i,i) = sqrt(nq); // conservative with the variance of ang vel meas.
     }
     _Estate(i,0) = 0.0; 
   }
 }
 else {
   // initialize for UKF
   _udt = dt;
   _uspd = spd;
   // we can set the measurment noise covariance matrix here (at least that can
   // stay constant
   _UR.setZeros();
   _UP.setZeros();
   for (int i = 0; i < 7; ++i) {
     if (i < 4) {
       _UP(i,i) = 1.0;
       _UR(i,i) = nq;
     }
     else {
       _UP(i,i) = INITVAR;
       _UR(i,i) = sqrt(nq); // conservative with the variance of ang vel meas.
     }
     _Estate(i,0) = 0.0; 
   }
 }

}


// do kalman filter and prediciton

Pose *
KalmanPVMMV::filter(Pose *userpos, double pinterval) 
{
 ++_ctr;

 if (_ctr == 1) { // initialize the state vectors and return the orig pose
                  // for both KF and EKF and UKF
   for (int i = 0; i < 3; ++i) {
    (_state[i])(0,0) = userpos->_pos[i];
    //  if (i == 0)
    //   cerr << " " << _state[i](0,0) << " " << sqrt(_PP[i](0,0)) << endl;
   }

   if (!_ukfflag) {
     for (int i = 0; i < 3; ++i)
       _Estate(i,0) = userpos->_orient.V()[i];
     _Estate(3,0) = userpos->_orient.S();
     //   cerr << " " << _Estate(0,0) << " " << sqrt(_EP(0,0)) << endl; 
   }
   else {
     for (int i = 0; i < 3; ++i)
       _Ustate(i,0) = userpos->_orient.V()[i];
     _Ustate(3,0) = userpos->_orient.S();
     //   cerr << " " << _Ustate(0,0) << " " << sqrt(_UP(0,0)) << endl; 
   }
  return userpos;
 }

 // do the kalman filter for position only -- Orientation to come
 kalmanFilter(userpos, 2, 2, kPVMMV);

 
 // we have three new state vectors 
 Vec3d ppos(0.0,0.0,0.0);

 if (_ctr < MAXBEGIN) // let it run to stabilize before we predict
   ;
 else {
   // otherwise do the prediction
   // create the prediction matrix
   MMatrix pmat = _fundMat;
   pmat(0,1) = pinterval;
  
   for (int i = 0; i < 3; ++i) 
     ppos[i] = (pmat *_state[i])(0,0); 
 }

  ////////////////////////////////////////////////////////
  // here we do the EKF or UKF work for orientation
  if (!_ukfflag)
    extendedKalmanFilter(userpos, 7, 7, kPVMMV);
  else 
    unscentedKalmanFilter(userpos, 0, kPVMMV);


   // we have the new state vector for orientation
   if (_ctr < MAXBEGIN) // let it run to stabilize before we predict
    return userpos;
   
   // predict
   if (!_ukfflag) {
     Quat prot(0,0,0,1);
     Quat tq(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0));
     Quat tw(_Estate(4,0), _Estate(5,0), _Estate(6,0), 0);
     prot = tq + (pinterval * 0.5 * (tq * tw)); 
     return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
   }
   else {
     Quat prot(0,0,0,1);
     Quat tq(_Ustate(0,0), _Ustate(1,0), _Ustate(2,0), _Ustate(3,0));
     Quat tw(_Ustate(4,0), _Ustate(5,0), _Ustate(6,0), 0);
     prot = tq + (pinterval * 0.5 * (tq * tw)); 
     return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot));
   }
}




